/**
 * Driver program for the Steganography assignment.  This program creates the image and drawing
 * canvas, responds to button events, does image conversion, and responds to user events.
 * 
 * @author Keith Schwarz (htiek@cs.stanford.edu)
 */

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

import javax.imageio.*;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;

import java.io.*;
import java.util.*;

import acm.program.*;
import acm.util.*;
import acm.gui.*;
import acm.graphics.*;

public class Steganography extends Program implements SteganographyConstants {
	/* Horizontal/vertical gap between images. */
	private static final int HGAP = 5;
	private static final int VGAP = 5;
	
	/* Recommended application width, which should be big enough to hold both images. */
	public static final int APPLICATION_WIDTH = 900;
	
	/* Valid file extensions for image types that we can write. */
	private static final String[] SAVE_IMAGE_EXTENSIONS = new String[] { ".png", ".bmp", ".wbmp" } ;
	
	/* Valid file extensions for images that we can read. */
	private static final String[] LOAD_IMAGE_EXTENSIONS = new String[] {
		".png", ".bmp", ".wbmp", ".jpg", ".gif", ".jpeg"
	};
	
	/* The actual drawing and picture canvas. */
	private SteganographyDrawingCanvas drawing = new SteganographyDrawingCanvas();
	private SteganographyPictureCanvas picture = new SteganographyPictureCanvas();
	
	public void init() {
		/* Disable caching so if we rewrite a file, we see the right version. */
		MediaTools.setCachingEnabled(false);
		
		/* Set up the display. */
		setLayout(new TableLayout(3, 2, HGAP, VGAP));
		add(new JLabel("Secret Drawing"));
		add(new JLabel("Master Image"));
		add(drawing);
		add(picture);
		add(new JButton("Clear Drawing"));
		add(new JButton("Choose Image"));
		
		add(new JButton("Hide Message"), SOUTH);
		add(new JButton("Recover Message"), SOUTH);
		
		/* Ask to receive events. */
		addActionListeners();
	}
		
	/**
	 * Responds to interactor events.
	 */
	public void actionPerformed(ActionEvent e) {
		if (e.getActionCommand().equals("Clear Drawing")) {
			drawing.clear();
		} else if (e.getActionCommand().equals("Choose Image")) {
			chooseImage();
		} else if (e.getActionCommand().equals("Hide Message")) {
			hideMessage();
		} else if (e.getActionCommand().equals("Recover Message")) {
			recoverMessage();
		}
	}
	
	/**
	 * Returns a JFileChooser that can read or write the specified file types.
	 * 
	 * @param extensions Legal file extensions
	 * @param description A textual description of those file types.
	 * @return A JFileChooser for those extensions
	 */
	private JFileChooser getFileChooser(final String[] extensions, final String description) {
		JFileChooser fc = new JFileChooser();
		fc.setFileFilter(new FileFilter() {
			@Override
			public boolean accept(File filename) {
				return filename.isDirectory() || Arrays.asList(extensions).contains(extensionOf(filename));
			}
			
			@Override
			public String getDescription() {
				return description;
			}
		});
		
		/* Set the working directory to the current images directory. */
		fc.setCurrentDirectory(new File("steganography"));
		
		return fc;
	}
	
	/**
	 * Prompts the user to choose which image they want to load, then
	 * loads that image.
	 */
	private void chooseImage() {
		/* Get a file chooser capable of choosing images. */
		JFileChooser fc = getFileChooser(LOAD_IMAGE_EXTENSIONS, "Image files");
		if (fc.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
			try {
				/* Try setting the image.  This might fail if we get a bad file. */
				picture.setImage(new GImage(fc.getSelectedFile().getAbsolutePath()));
			} catch (ErrorException ex) {
				getDialog().showErrorMessage("Could not open that image.");
			}
		}
	}
	
	/**
	 * Hides the secret message, optionally prompting the user to save.
	 */
	private void hideMessage() {
		try {
			/* Run the steganography algorithm to hide the message in the picture. */
			GImage image = SteganographyLogic.hideMessage(toBooleanArray(getImage(drawing)), getImage(picture));
			if (image == null)
				throw new NullPointerException("Image is null.");
			
			picture.setImage(image);
		
			/* Ask whether to save. */
			if (getDialog().readBoolean("Message hidden.  Do you want to save?", "Yes", "No")) {
				JFileChooser fc = getFileChooser(SAVE_IMAGE_EXTENSIONS, ".png, .bmp, and .wbmp files");
				if (fc.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
					saveImage(picture.getImage(), fc.getSelectedFile());
				}
			}
		} catch (Exception e) {
			getDialog().showErrorMessage("An error occurred: " + e.getMessage());
		}
	}
	
	/**
	 * Recovers the secret message and displays it in the drawing canvas.
	 */
	private void recoverMessage() {
		try {
			drawing.setImage(toGImage(SteganographyLogic.findMessage(getImage(picture))));
		} catch (Exception e) {
			getDialog().showErrorMessage("An error occurred: " + e.getMessage());
		}
	}
	
	/**
	 * Given a GImage of black and white pixels, returns a boolean array formed by taking that
	 * image and treating each value as true or false.
	 * 
	 * @param image The image to convert.
	 * @return A boolean matrix from that image.
	 */
	private boolean[][] toBooleanArray(GImage image) {
		/* Get the underlying pixel array. */
		int[][] pixels = image.getPixelArray();
		
		/* Cache its size for simplicity. */
		int numRows = pixels.length;
		int numCols = pixels[0].length;
		
		/* Create a white pixel to use as a reference. */
		int whitePixel = GImage.createRGBPixel(255, 255, 255);
		
		boolean[][] result = new boolean[numRows][numCols];
		for (int row = 0; row < numRows; row++) {
			for (int col = 0; col < numCols; col++) {
				/* Store whether the pixel is black. */
				result[row][col] = (pixels[row][col] != whitePixel);
			}
		}
		
		return result;
	}
	
	/**
	 * Given a boolean array, converts it to a monochrome GImage.
	 * 
	 * @param pixels The array of pixels to convert.
	 * @return The resulting GImage.
	 */
	private GImage toGImage(boolean[][] pixels) {
		/* Confirm this is a valid matrix. */
		if (pixels == null) throw new NullPointerException("Returned array is null.");
		if (pixels.length != CANVAS_HEIGHT) throw new IllegalArgumentException("Incorrect size for returned array.");
		for (int i = 0; i < pixels.length; i++) {
			if (pixels[i] == null)
				throw new IllegalArgumentException("Inner array is null.");
			if (pixels[i].length != CANVAS_WIDTH) throw new IllegalArgumentException("Incorrect size for returned array.");
		}
		
		/* Build up an image of pixels whose color depends on the given booleans. */
		int[][] result = new int[CANVAS_HEIGHT][CANVAS_WIDTH];
		for (int row = 0; row < pixels.length; row++) {
			for (int col = 0; col < pixels[row].length; col++) {
				result[row][col] = pixels[row][col]? GImage.createRGBPixel(0, 0, 0) : GImage.createRGBPixel(255, 255, 255);
			}
		}
		
		return new GImage(result);
	}

	/**
	 * Given a filename, returns the extension of that filename, or "" if it does not
	 * have one.
	 * 
	 * @param filename The filename in question.
	 * @return Its extension, or "" if it has none.
	 */
	private String extensionOf(File filename) {
		/* Find the suffix; fail if none is found. */
		int lastDot = filename.getName().lastIndexOf('.');
		if (lastDot == -1) return "";
		
		/* Determine the extension. */
		return filename.getName().substring(lastDot);
	}
	
	/**
	 * Returns whether a given file extension is a valid file extension we can use
	 * when saving files.
	 * 
	 * @param extension The extension in question.
	 * @return Whether it is a valid file extension.
	 */
	private boolean isLegalExtension(String extension) {
		/* Confirm it's one of .png, .bmp, or .wbmp, which are supported. */
		for (String legal: SAVE_IMAGE_EXTENSIONS) {
			if (legal.equals(extension)) return true;
		}
		return false;
	}
	
	/**
	 * Given an image, converts that image into a BufferedImage.
	 * <p>
	 * This code is based on code taken from this source:
	 * <p>
	 * http://www.javareference.com/jrexamples/viewexample.jsp?id=112
	 * 
	 * @param img The image in question
	 * @return A BufferedImage containing the contents of that image.
	 */
	private BufferedImage getBufferedImageFromImage(Image img) { 
        /* Force the image to be loaded into memory. */
        img = new ImageIcon(img).getImage(); 
         
        /* Create a BufferedImage large enough to hold the result. */ 
        BufferedImage bufferedImage = new BufferedImage(CANVAS_WIDTH, CANVAS_HEIGHT, BufferedImage.TYPE_INT_RGB); 

        /* Get a graphics context for the new BufferedImage. */
        Graphics g = bufferedImage.createGraphics(); 

        /* Draw the image onto the BufferedImage. */ 
        g.drawImage(img, 0, 0, null); 
         
        /* Clean up the graphics context. */
        g.dispose(); 
         
        return bufferedImage; 
    } 
	
	/**
	 * Given a GImage and a file, writes that GImage to the file.
	 * 
	 * @param image The image to write.
	 * @param file Where it should be written.
	 */
	private void saveImage(GImage image, File file) {
		try {
			/* Find the extension, if one exists.  If not, pretend it's a png. */
			int lastDot = file.getName().lastIndexOf('.');
			if (lastDot == -1) {
				file = new File(file.getAbsolutePath() + ".png");
				System.out.println(file);
			}
			
			/* Get that extension. */
			String extension = extensionOf(file.getAbsoluteFile());
			if (!isLegalExtension(extension)) throw new ErrorException("Unsupported file format.");
			
			/* Turn the GImage into a BufferedImage. */
			BufferedImage bufferedImage = getBufferedImageFromImage(MediaTools.createImage(image.getPixelArray()));
			
			/* Output to that file.  The extension to use should first be trimmed of its leading dot. */
			ImageIO.write(bufferedImage, extension.substring(1), file);
		} catch (ErrorException e) {
			getDialog().showErrorMessage(e.getMessage());
		} catch (IOException e) {
			getDialog().showErrorMessage("An error occurred saving the image.");
		}
	}
	
	/**
	 * Given a component, returns a GImage holding the contents of that component.
	 * Source: http://stackoverflow.com/questions/1349220/convert-jpanel-to-image
	 * 
	 * @param component The component to draw.
	 * @return image A GImage of that component.
	 */
	private GImage getImage(Component c) {
		/* Create a BufferedImage large enough to hold the component. */
	    BufferedImage bi = new BufferedImage(CANVAS_WIDTH, CANVAS_HEIGHT, BufferedImage.TYPE_INT_RGB);
	    
	    /* Create a graphics context and render the image into that context. */
	    Graphics2D g = bi.createGraphics();
	    c.paint(g);
	    
	    return new GImage(bi);
	}
}
