/**
 * Main program for the Tone Matrix assignment.  This program is responsible for setting up
 * the display, interacting with buttons, and saving/loading matrices.
 * 
 * @author Keith Schwarz (htiek@cs.stanford.edu)
 */

import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;

import acm.program.*;
import acm.util.*;

public class ToneMatrix extends Program {
	/* The displayed tone matrix in the center of the program. */
	private ToneMatrixDisplay display = new ToneMatrixDisplay();
	
	/* The size of the tone matrix.  If you want to change this value, you should update the
	 * value in ToneMatrixConstants, rather than changing this.
	 */
	private static final int MATRIX_SIZE = ToneMatrixConstants.size();
	
	/**
	 * Creates the window and attaches a tone matrix.
	 */
	public void init() {
		add(display);
		
		/* Create the control buttons, then indicate that we want to receive
		 * events.
		 */
		add(new JButton("Clear"), SOUTH);
		add(new JButton("Save Matrix"), SOUTH);
		add(new JButton("Load Matrix"), SOUTH);
		addActionListeners();
	}
	
	/**
	 * Reacts to button presses.
	 */
	public void actionPerformed(ActionEvent e) {
		if (e.getActionCommand().equals("Clear")) {
			display.clear();
		} else if (e.getActionCommand().equals("Save Matrix")) {
			saveMatrix();
		} else if (e.getActionCommand().equals("Load Matrix")) {
			loadMatrix();
		}
	}
	
	/**
	 * Returns the extension of a file, or the empty string if there is no extension.
	 * 
	 * @param filename The file whose extension should be extracted.
	 * @return The extension of that file, or "" if it has none.
	 */
	private String extensionOf(File filename) {
		/* Find the suffix; fail if none is found. */
		int lastDot = filename.getName().lastIndexOf('.');
		if (lastDot == -1) return "";
		
		/* Determine the extension. */
		return filename.getName().substring(lastDot);
	}
	
	/**
	 * Returns a JFileChooser that can choose .matrix files from the current
	 * directory.
	 * 
	 * @return A JFileChooser for load/save dialogs.
	 */
	private JFileChooser getFileChooser() {
		JFileChooser fc = new JFileChooser();
		fc.setFileFilter(new FileFilter() {
			@Override
			public boolean accept(File filename) {
				return filename.isDirectory() || extensionOf(filename).equals(".matrix");
			}
			
			@Override
			public String getDescription() {
				return ".matrix Files";
			}
		});
		
		/* Set the current directory to be the "matrices" directory so that the
		 * files we see are reasonable.
		 */
		fc.setCurrentDirectory(new File("matrices"));
		
		return fc;
	}
	
	/**
	 * Prompts the user for a filename, then saves the matrix to that location.
	 */
	private void saveMatrix() {
		/* Create a file chooser that asks for people to choose files with
		 * the .matrix extension.
		 */
		JFileChooser fc = getFileChooser();
		
		/* Wait for the user to pick a file, and if they do, write the matrix
		 * to disk.
		 */
		if (fc.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
			/* Tack on .matrix as an extension if it's not already present. */
			String filename = fc.getSelectedFile().getAbsolutePath();
			if (extensionOf(fc.getSelectedFile()).equals("")) {
				filename += ".matrix";
			}
			
			/* Try to write the file. */
			try {
				writeMatrixToFile(display.getMatrix(), filename);
			} catch (ErrorException err) {
				getDialog().showErrorMessage("Error saving file: " + err.getMessage());
			}
		}
	}
	
	private void loadMatrix() {
		/* Create a file chooser that asks for people to choose files with
		 * the .matrix extension.
		 */
		JFileChooser fc = getFileChooser();
		
		if (fc.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
			try {
				display.setMatrix(readMatrixFromFile(fc.getSelectedFile().getAbsolutePath()));
			} catch (ErrorException err) {
				getDialog().showErrorMessage("Error loading file: " + err.getMessage());
			}
		}
	}
	
	/**
	 * Saves the contents of the tone matrix to the indicated file.
	 * <p>
	 * The tone matrix should be saved to disk by writing out the first row of booleans, one
	 * per line, then the second row of booleans, then the third, etc.
	 * <p>
	 * If you encounter an error when saving the file, you should throw an ErrorException so
	 * that the program will display the appropriate error message.
	 * 
	 * @param toneMatrix The tone matrix to save.
	 * @param fileName The name of the file to save it to.
	 */
	public void writeMatrixToFile(boolean[][] toneMatrix, String fileName) {
		try {
			PrintWriter pr = new PrintWriter(new FileWriter(fileName));
			
			/* Write the file in row-major order - each row is printed one at a time, from
			 * the top to the bottom.
			 */
			for (int row = 0; row < MATRIX_SIZE; row++) {
				for (int col = 0; col < MATRIX_SIZE; col++) {
					pr.println(toneMatrix[row][col]);
				}
			}
			pr.close();
		} catch (IOException e) {
			throw new ErrorException(e);
		}
	}
	
	/**
	 * Loads the tone matrix stored in the indicated file.
	 * <p>
	 * The entries in the tone matrix are stored one row at a time, with each row written out as a list
	 * of booleans, one per line.
	 * <p>
	 * If you encounter an error when reading the tone matrix, you should throw an ErrorException to
	 * indicate this.
	 * 
	 * @param fileName The name of the file to read. 
	 * @return The tone matrix stored in that file.
	 */
	public boolean[][] readMatrixFromFile(String fileName) {
		try {
			boolean[][] result = new boolean[MATRIX_SIZE][MATRIX_SIZE];
			BufferedReader br = new BufferedReader(new FileReader(fileName));
			
			/* Read the file in row-major order - each row is printed one at a time, from
			 * the top to the bottom.
			 */
			for (int row = 0; row < MATRIX_SIZE; row++) {
				for (int col = 0; col < MATRIX_SIZE; col++) {
					result[row][col] = Boolean.parseBoolean(br.readLine());
				}
			}
			br.close();
			return result;
		} catch (IOException e) {
			throw new ErrorException(e);
		}
	}
	
	public void run() {
		/* Grab the samples we're going to use. */
		double[][] samples = ToneMatrixWaveGenerator.getSamples();
		
		while (true) {
			for (int i = 0; i < MATRIX_SIZE; i++) {
				/* Get a snapshot of the tone matrix. */
				boolean[][] matrix = display.getMatrix();
				
				/* Hightlight the current column. */
				display.highlightColumn(i);
				
				/* Convert it to a string. */
				double[] toPlay = ToneMatrixLogic.matrixToMusic(matrix, i, samples);
				
				/* Play that music string. */
		        StdAudio.play(toPlay);
			}
		}
	}
}
