/**
 * Code to generate the waveforms associated with each of the notes in the
 * tone matrix.
 * <p>
 * The generated sound samples are produced by taking a normal sine wave, then
 * easing into the waveform and out of the waveform.
 * 
 * @author Keith Schwarz (htiek@cs.stanford.edu)
 */

import acm.graphics.*;

/**
 * A class that generates waveforms that can be played by the tone matrix.
 */
public class ToneMatrixWaveGenerator {
	/* The five basic frequencies used in the tone matrix are taken from the pentatonic scale. */
	private static final double[] FREQUENCIES = {
		130.81,  // C
		146.83,  // D
		164.81,  // E
		196.00,  // G
		220.00,  // A
	};
	
	/**
	 * Given a frequency, returns the length of one period of a wave at that frequency, represented
	 * in samples.
	 * 
	 * @param frequency The frequency whose period should be computed.
	 * @return The period of that frequency, in samples.
	 */
	private static int periodOf(double frequency) {
		return GMath.round(StdAudio.SAMPLE_RATE / frequency);
	}
	
	/**
	 * Generates samples for a sine wave of a given frequency.
	 * 
	 * @param frequency The frequency in question.
	 * @return An array of samples of a sine wave of that frequency.
	 */
	private static double[] sineWave(double frequency) {
		double[] result = new double[periodOf(frequency)];
		for (int i = 0; i < result.length; i++) {
			result[i] = Math.sin(2.0 * Math.PI * i / result.length);
		}
		return result;
	}

	/**
	 * Returns an array of sound samples corresponding to the notes in the tone matrix.
	 * <p>
	 * These samples are formed by starting with sine waves of the appropriate frequencies, then
	 * easing into the tone and out of the tone.  If you want to change the sounds that are being
	 * produced, you can do so by changing this code.
	 * 
	 * @return The samples to be played.
	 */
	public static double[][] getSamples() {
		/* Create the resulting sample array. */
		double[][] result = new double[ToneMatrixConstants.size()][];
		for (int note = 0; note < result.length; note++) {
			/* Determine which note is being played. */
			int noteIndex = note % FREQUENCIES.length;
			
			/* Generate a sine wave at that frequency. */
			double[] triangle = sineWave(FREQUENCIES[noteIndex] * Math.pow(2.0, 1 + note / FREQUENCIES.length));

			/* Replicate that wave as many times as is necessary to fill in the sample. */
			double[] generated = new double[ToneMatrixConstants.sampleSize()];
			for (int x = 0; x < generated.length; x++) {
				generated[x] = triangle[x % triangle.length];
			}

			/* Ease out of the note. */
			double attenuationFactor = -1.0 / generated.length;
			for (int x = 0; x < generated.length; x++) {
				generated[x] *= (1.0 + x * attenuationFactor);
			}

			/* Ease into the note. */
			double leadInFactor = 1.0 / (generated.length / 64);
			for (int x = 0; x < generated.length / 64; x++) {
				generated[x] *= leadInFactor * x * x / (generated.length / 64);
			}

			/* Store the generated note. */
			result[note] = generated;			
		}
		return result;
	}
}
